import attr
import pandas as pd
import os
import csv
from pathlib import Path
from urllib.parse import urlparse
import requests

BASE_PATH = os.getcwd()


def _dir_exists(path):
    return os.path.isdir(path) and os.path.exists(path)

def absolute_path(p: str) -> str:
    return os.path.join(os.path.dirname(os.path.abspath(__file__)), p)

@attr.s(hash=True)
class Dataset(object): 
    dfile = attr.ib()  # path name
    variables = attr.ib()  # list of Variable objects <-- TODO: may not need this in new implementation....
    pid_col_name = attr.ib()  # name of column in pandas DataFrame that has participant ids
    row_pids = attr.ib(init=False)  # list of unique participant ids
    data = attr.ib(init=False)  # pandas DataFrame
    
    @staticmethod
    def load(path: str, name):
        assert(isinstance(path, str))
        home = Path.home()
        tea_path = home / '.tea'
        if not _dir_exists(tea_path):
            os.mkdir(tea_path)
        data_path = tea_path / 'data'
        if not _dir_exists(data_path):
            os.mkdir(data_path)
        
        url = urlparse(path)
        csv_name = name if '.csv' in name else str(name + '.csv')
        csv_path = data_path / csv_name

        # URL
        if url.scheme != '':
            data = requests.get(path)

            with open(csv_path, 'w') as f:
                writer = csv.writer(f)
                reader = csv.reader(data.text.splitlines())

                for row in reader:
                    writer.writerow(row)
        else: 
            with open(path, 'r') as readfile: 
                reader = csv.reader(readfile)
                with open(csv_path, 'w') as writefile:
                    writer = csv.writer(writefile)

                    for row in reader:
                        writer.writerow(row)

        return csv_path

    def __attrs_post_init__(self):
        if isinstance(self.dfile, (str, Path)):
            # import pdb;pdb.set_trace()
            self.data = pd.read_csv(self.dfile)
        elif isinstance(self.dfile, pd.DataFrame):
            self.data = self.dfile
        else:
            raise ValueError(f"Dataset expected DataFrame, str, or Path. Received: {self.dfile}")

        # if self.pid_col_name:
        #     # Reindex DataFrame indices to be pids
        #     self.data.set_index(self.pid_col_name, inplace=True)
        # else: 
            # Treat each row as a unique observation

    def __getitem__(self, var_name: str):
        for v in self.variables:  # checks that the Variable is known to the Dataset object
            if v.name == var_name: 
                return self.data[var_name]  # returns the data, not the variable object

    # Returns variable object -- may want to altogether replace the __getitem__
    def get_variable(self, var_name: str): 
        for v in self.variables: 
            if v.name == var_name: 
                return v

    def get_variable_data(self, var_name: str):
        for v in self.variables: 
            if v.name == var_name:
                return {'dtype': v.dtype,
                        'categories': v.categories} 

    # SQL style select
    def select(self, col: str, where: list = None):

            # Add back ticks to variable names that have spaces in them 
        def format_where(vars: list, where: list): 
            formatted_where = []
            for clause in where: 
                bt_var_name = '`' # opening back tick
                for var in vars: 
                    # Is the variable in the where clause? 
                    if var.name in clause: 
                        # Does the variable name have a space in it? 
                        if ' ' in var.name: 
                            bt_var_name += var.name 
                            bt_var_name += '`' # closing back tick
                            split_clause = clause.split(var.name)
                            assert(len(split_clause) == 2)
                            assert(len(split_clause[0]) == 0)
                            formatted_where.append(bt_var_name + split_clause[1])
                        else: 
                            formatted_where.append(clause)
            
            formatted_where_str = "&".join(formatted_where)
            return formatted_where_str            

        df = self.data
        if where: # not None
            # query = build_query(self.variables, where)
            query = format_where(self.variables, where)
            # import pdb; pdb.set_trace()
            res = df.query(query)[col] # makes a copy
            # import pdb; pdb.set_trace()
        else: 
            res = df[col]

        return res
    
    def has_empty_data(self):
        assert(self.data is not None)
        assert(isinstance(self.data, pd.DataFrame))

        return self.data.empty


from typing import Dict

from tea.global_vals import dv_identifier, iv_identifier, \
    contributor_identifier, experiment_identifier, observational_identifier, outcome_identifier, study_type_identifier


class StudyTypeDeterminer:
    def determine_study_type(self, vars_data: list, design: Dict[str, str]):
        if design:
            # Is the study type explicit? If so...
            if study_type_identifier in design:
                # Is this study an experiment or observational study?
                if design[study_type_identifier] == experiment_identifier:
                    return experiment_identifier
                # Is this study an observational study?
                elif design[study_type_identifier] == observational_identifier:
                    return observational_identifier
                # We don't know what kind of study this is.
                else:
                    raise ValueError(f"Type of study is not supported:{design[study_type_identifier]}. Is it an experiment or an observational study?")
            # The study type is not explicit, so let's check the other properties...
            else:
                # This might be an experiment.
                if iv_identifier in design and dv_identifier in design:  # dv_identifier??
                    return experiment_identifier
                elif contributor_identifier in design and outcome_identifier in design:
                    return observational_identifier
                # We don't know what kind of study this is.
                else:
                    raise ValueError(f"Type of study is not supported:{design}. Is it an experiment or an observational study?")



from typing import Union

import pandas as pd

from tea.runtimeDataStructures.dataset import Dataset
from tea.ast import (Variable, DataType, Literal, Relate, Relationship)

from collections import OrderedDict
from pathlib import Path

iv_identifier = 'independent variable'
dv_identifier = 'dependent variable'
var_identifier = 'variable'
categorical_prediction = 'categorical prediction'
continuous_prediction = 'continuous prediction'
categorical_prediction_delimiter = ':'
continuous_prediction_delimiter = '~'
categorical_comparators = ['>', '<', '==', '!=']
continuous_positive_relationship = '+'
continuous_negative_relationship = '-'


def const(val):
    return Literal(val)


def ordinal(var_name: str, ordered_categories: list):
    # Create order tuple
    categories = OrderedDict()
    for i, c in enumerate(ordered_categories):
        categories[c] = i + 1  # start at 1 not 0

    return Variable.from_spec(var_name, DataType.ORDINAL, categories, [1, len(categories)])


def isordinal(var: Variable):
    return var.dtype is DataType.ORDINAL


def nominal(var_name: str, unordered_categories: list):
    categories = OrderedDict()
    for i, c in enumerate(unordered_categories):
        categories[c] = -1
    return Variable.from_spec(var_name, DataType.NOMINAL, categories, None)


def isnominal(var: Variable):
    return var.dtype is DataType.NOMINAL


def interval(var_name: str, drange: list = None):
    return Variable.from_spec(var_name, DataType.INTERVAL, None,
                              drange)  # treat range like categories, check that all values are within range


def isinterval(var: Variable):
    return var.dtype is DataType.INTERVAL


def ratio(var_name: str, drange: list = None):
    return Variable.from_spec(var_name, DataType.RATIO, None,
                              drange)  # treat range like categories, check that all values are within range


def isratio(var: Variable):
    return var.dtype is DataType.RATIO


def isnumeric(var: Variable):
    return isratio(var) or isinterval(var)


def iscategorical(var: Variable):
    return isnominal(var) or isordinal(var)


# @param pid is the name of the column with participant ids
def load_data(source_name: Union[str, Path, pd.DataFrame], vars: list, pid: str):
    return Dataset(source_name, vars, pid)


def load_data_from_url(url: str, name: str):
    return Dataset.load(url, name)


def select(var: Variable, op: str, other: Literal):
    if op == '==':
        return var.subset_equals(other)
    elif op == '!=':
        return var.subset_not_equals(other)
    elif op == '<':
        return var.subset_lt(other)
    elif op == '<=':
        return var.subset_le(other)
    elif op == '>':
        return var.subset_gt(other)
    elif op == '>=':
        return var.subset_ge(other)
    else:
        raise ValueError(f"Do not support the operator{op}")


# @param vars: list of all vars in tuple form (Variable, 'iv/dv/variable')
# @param role: "role" of Variable (i.e., independent variable, dependent variable, variable)
# @returns new list of vars with specified role
# def get_vars(role: str, vars: list): 
#     vars_role = []
#     for v,r in vars:
#         if r == role: 
#             vars_role.append(v)

#     return vars_role

# TODO: need to check that the prediction is well-formed (VALUES that are ordered exist, for example)
"""
    Grammar for predictions: 
    # Cat x Numeric
    # what matters is that the var using to form groups in comparison is nominal/categorical 
    <, >, ==, != --> Tests null hypothesis that there is no difference OR ask for one-sidedness in tests (as constraint?)
    predict(groups: Variable, outcome: Variable, prediction: str=None)
    # Cat x Cat
    Frequency(??)
    predict(groups: Variable, outcome: Variable, prediction: str=None)
    # Numeric x Cat
    # we need to know which variable is factor vs. predictor
    logistic regression
    predict(var: Variable, outcome: Variable, prediction: str=None)
    # Numeric x Numeric --> Should allow for both
    # Need to know which variable is factor vs predictor
    Prediction: y ~ - var + var + var (e.g., time ~ + condition + age || prediction: condition:a > condition:b)
    Prediction: + var, - var
    predict(outcome: Variable, factors: list of variables, prediction: str=None)
    predict(factors: list, outcome: Variable, prediction: str=None)
    --> check type of outcome
    --> check number of factors: if > 1 then multivariate case else bivariate case, check for factor type
    --> in multivariate case, the types of factors don't have a bearing on the ....?
"""


# def is_var_in_list(var_name: str, vars: list):
#     names = [v.name for v in vars]

#     return var_name in names

def get_var_from_list(var_name: str, vars: list):
    for v in vars:
        if v.name == var_name:
            return v

    return None  # Does not exist


def is_well_formed_prediction(prediction_type: str, vars: list, prediction: str):
    # import pdb; pdb.set_trace()
    if categorical_prediction == prediction_type:
        var_ind = prediction.index(categorical_prediction_delimiter)
        var_name = prediction[:var_ind].strip()

        # Does the variable actually exist in the list of variables?
        var = get_var_from_list(var_name, vars)
        return var is not None
        # if var:
        #     # Does the category exist in the var? (implictly checks that var is a categorical variable)
        #     return var.category_exists(category)
        # else: 
        #     return ValueError(f"The variable used in prediction{var} is not in list of variables comparing{vars}")
    elif continuous_prediction == prediction_type:

        # Prediction includes categorical variable value
        if categorical_prediction_delimiter in prediction:
            cat_delimiter_ind = prediction.index(categorical_prediction_delimiter)
            num_delimiter_ind = prediction.index(continuous_prediction_delimiter)

            categorical_var_name = prediction[:cat_delimiter_ind]
            # categorical_var = get_var_from_list(categorical_var_name, vars)

            # LHS is categorical
            if cat_delimiter_ind < num_delimiter_ind:
                lhs = categorical_var_name.strip()
                category = lhs[cat_delimiter_ind + 1:num_delimiter_ind]
                lhs_var = get_var_from_list(lhs, vars)
                assert (lhs_var.category_exists(category))
                rhs = prediction[num_delimiter_ind + 1:].strip()
            else:
                assert (cat_delimiter_ind > num_delimiter_ind)
                lhs = prediction[:num_delimiter_ind].strip()
                rhs = prediction[num_delimiter_ind + 1:cat_delimiter_ind].strip()
                category = rhs[cat_delimiter_ind + 1:].strip()
                rhs_var = get_var_from_list(rhs, vars)
                assert (rhs_var.category_exists(category))

        # Prediction does not include categorical variable value (can still have a categorical variable)
        else:
            delimiter_ind = prediction.index(continuous_prediction_delimiter)
            lhs = prediction[:delimiter_ind].strip()
            rhs = prediction[delimiter_ind + 1:].strip()

        if continuous_negative_relationship in lhs:
            lhs = lhs[lhs.index(continuous_negative_relationship) + 1:].strip()
        elif continuous_positive_relationship in lhs:
            lhs = lhs[lhs.index(continuous_negative_relationship) + 1:].strip()
        else:
            pass

        if continuous_negative_relationship in rhs:
            rhs = rhs[rhs.index(continuous_negative_relationship) + 1:].strip()
        elif continuous_positive_relationship in rhs:
            rhs = rhs[rhs.index(continuous_positive_relationship) + 1:].strip()
        else:
            pass

        return get_var_from_list(lhs, vars) and get_var_from_list(rhs, vars)
    else:
        raise ValueError(f"Malformed prediction. Is neither categorical nor numeric:{prediction}")


def create_prediction(prediction_type: str, vars: list, prediction: str):
    global categorical_prediction, categorical_prediction_delimiter, categorical_comparators
    global continuous_prediction, continuous_prediction_delimiter, positive_relationship, negative_relationship

    if categorical_prediction == prediction_type:
        # comparator = None

        for c in categorical_comparators:
            if c in prediction:
                delimiter_ind = prediction.index(categorical_prediction_delimiter)
                comparator_ind = prediction.index(c)

                var_name = prediction[:delimiter_ind].strip()
                var = get_var_from_list(var_name, vars)
                lhs = prediction[delimiter_ind + 1:comparator_ind].strip()
                # Add len(c) to index in case comparator is longer than one character
                rhs = prediction[comparator_ind + len(c):].strip()

                types = set(type(k) for k in var.categories.keys())
                assert (len(types) == 1)  # Assert all keys have the same type
                cat_type = next(iter(types))

                if isinstance(lhs, str) and cat_type == int:
                    lhs = int(lhs)
                    rhs = int(rhs)

                assert (lhs in var.categories.keys())
                assert (rhs in var.categories.keys())
                if c == '<':
                    return [const(lhs) < const(rhs)]
                elif c == '>':
                    return [const(lhs) > const(rhs)]
                elif c == '==':
                    return [const(lhs) == const(rhs)]
                else:
                    assert (c == '!=')
                    return [const(lhs) != const(rhs)]
    else:
        assert (continuous_prediction == prediction_type)

        # Prediction includes categorical variable value
        if categorical_prediction_delimiter in prediction:
            cat_delimiter_ind = prediction.index(categorical_prediction_delimiter)
            num_delimiter_ind = prediction.index(continuous_prediction_delimiter)

            categorical_var_name = prediction[:cat_delimiter_ind]
            # categorical_var = get_var_from_list(categorical_var_name, vars)

            # LHS is categorical
            if cat_delimiter_ind < num_delimiter_ind:
                lhs = categorical_var_name.strip()
                rhs = prediction[num_delimiter_ind + 1:].strip()
            else:
                assert (cat_delimiter_ind > num_delimiter_ind)
                lhs = prediction[:num_delimiter_ind].strip()
                rhs = prediction[num_delimiter_ind + 1:cat_delimiter_ind].strip()

        # Prediction does not include categorical variable value (can still have a categorical variable)
        else:
            delimiter_ind = prediction.index(continuous_prediction_delimiter)
            lhs = prediction[:delimiter_ind].strip()
            rhs = prediction[delimiter_ind + 1:].strip()

            lhs_dir = '+'  # Positive relationship/change in variable is the default
            rhs_dir = '+'  # Positive relationship/change in variable is the default

            if lhs.startswith(continuous_negative_relationship) or lhs.startswith(continuous_positive_relationship):  # "as LHS decreases,..."
                lhs_dir = lhs[0]
                lhs = lhs[1:]
                lhs_var = get_var_from_list(lhs, vars)
            else:
                # do nothing to lhs, lhs_dir
                lhs_var = get_var_from_list(lhs, vars)

            if rhs.startswith(continuous_negative_relationship) or rhs.startswith(continuous_positive_relationship):
                rhs_dir = rhs[0]
                rhs = rhs[1:]
                rhs_var = get_var_from_list(rhs, vars)
            else:
                # do nothing to rhs, rhs_dir
                rhs_var = get_var_from_list(rhs, vars)

            if lhs_dir == rhs_dir:  # handles +lhs, +rhs and -lhs,
                return Relationship(lhs_var).positive(Relationship(rhs_var))
            else:  # handles +lhs, -rhs and -lhs, +rhs
                return Relationship(lhs_var).negative(Relationship(rhs_var))


# def predict(factors: list, outcome: Variable, prediction: str=None):
def predict(vars: list, predictions: list = None):
    formulated_predictions = []

    # Validate well-formedness of predictions
    if predictions:
        # Check that each prediction is well-formed
        for p in predictions:
            assert (isinstance(p, str))
            # Does the prediction pertain to categorical data?
            if categorical_prediction_delimiter in p and continuous_prediction_delimiter not in p:
                assert (is_well_formed_prediction(categorical_prediction, vars, p))
                pred = create_prediction(categorical_prediction, vars, p)

            # Does the prediction pertain to numerical data?
            elif continuous_prediction_delimiter in p:
                assert (is_well_formed_prediction(continuous_prediction, vars, p))
                # import pdb; pdb.set_trace()
                pred = create_prediction(continuous_prediction, vars, p)

            # Prediction pertains to neither categorical nor numerical data                 
            else:
                raise ValueError(f"Prediction is malformed: {p}")

            formulated_predictions.append(pred)

    return formulated_predictions


# Generic interface for observing relationships among variables (observational studies and experiments)
# @params: vars should be a list of Variables
def relate(vars: list, prediction: list = None):
    return Relate(vars, predict(vars, prediction))


# @params: iv could be the list of variables/groups want to compare on dv -
# may only want to compare 2 groups, not all conditions
# def compare(var_1: Variable, var_2: Variable, prediction: str=None, when: str=None) :

# def compare(iv: Variable, dv: Variable, prediction: str=None, when: str=None) :
def compare(vars: list, prediction: list = None, when: str = None):
    return Relate(vars, predict(vars, prediction))

    # if (isinstance(iv, Variable)):
    #     vars = [(iv, iv_identifier), (dv, dv_identifier)]
    #     if isnominal(iv) or isordinal(iv):
    #         return Relate(vars, predict(vars, prediction))
    #     elif isnumeric(iv):
    #         return Relate(vars, predict(vars, prediction))
    #     else: 
    #         raise ValueError(f"Invalid Variable type: {iv.dtype}")
    # elif (isinstance(iv, list)): 
    #     vars = []
    #     for i in iv: 
    #         vars.append((i, iv_identifier))
    #     return Relate(vars, predict(vars, prediction))
    # else: 
    #     raise ValueError (f"IV (first parameter) is not a Variable or list of Variables: {f.type}")

# def compare(var, var, semantically_same=True): # Compare two groups
#     pass

# compare(iv, dv, predictions)
# compare(dv, dv, groups=False) or compare (iv, iv, groups=False) 
# # Are they the same groups? 

# strategy = nominal('strategy', ['forking', 'caching', 'naive'])
# ratio('time', drange=[0, 10000], comes_from=(strategy, ['forking', 'caching', 'naive']))




from dataclasses import dataclass
from tea.logging.logging_target import LoggingTarget
from typing import Optional
import logging


@dataclass
class TeaLoggerConfiguration:
    logging_level: int = logging.INFO
    logging_target: LoggingTarget = LoggingTarget.STDOUT
    logging_file_target: Optional[str] = None



import attr
from enum import Enum
from typing import Union


class Node(object): 
    def relate(self, other):
        return Relate(self, other)

    # def compare(self, other):
    #     return Compare(self, other)
    
    def select(self, other):
        return Select(self, other)

    def __add__(self, other):
        return Add(self, other)

    def __sub__(self, other):
        return Sub(self, other)

    def __mul__(self, other):
        return Mul(self, other)

    def __truediv__(self, other):
        return Div(self, other)
    
    # MAY NOT NEED TO OVERRIDE AND, OR
    def __and__(self, other):
        return And(self, other)
    
    def __or__(self, other):
        return Or(self, other)


@attr.s(repr=False)
class DataType(Enum):  
    ORDINAL = 0
    NOMINAL = 1
    # for CALCULATIONS, INTERVAL vs. RATIO data is not important distinction
    # for INTERPRETATIONS, important distinction (?)
    INTERVAL = 2 
    RATIO = 3


@attr.s(hash=True, repr=False, cmp=False)
class Variable(Node): 
    # meta data about the variable, determines how the operators are to be/can be executed
    name = attr.ib()
    dtype = attr.ib(type=DataType)
    categories = attr.ib()
    drange = attr.ib()
    
    # children from which self/this Variable is "derived"
    rhs = attr.ib(type=Node)
    lhs = attr.ib(type=Node)

    @classmethod
    def from_spec(cls, name: str, dtype: DataType, cat: list=None, drange: list=None, rhs: Node=None, lhs: Node=None):
        return cls(name, dtype, cat, drange, rhs, lhs)
    
    def category_exists(self, category: str):
        return self.categories and category in self.categories
    
    def subset_equals(self, other: Node):
        return Equal(self, other)
    
    def subset_not_equals(self, other: Node):
        return NotEqual(self, other)

    def subset_lt(self, other: Node):
        return LessThan(self, other)

    def subset_le(self, other: Node):
        return LessThanEqual(self, other)
    
    def subset_ge(self, other: Node):
        return GreaterThanEqual(self, other)
    
    def subset_gt(self, other: Node):
        return GreaterThan(self, other)

    def __repr__(self): 
        return self.name


@attr.s(hash=True, repr=False)
class Select(Node):
    var: Variable # variable to filter on
    condition: tuple # inclusive bounding on both sides
    
    def __repr__(self):
        return (f"Select {self.var} on [{self.condition}]")


@attr.s(hash=True, repr=False)
class Equal(Node):
    lhs = attr.ib(type=Node)
    rhs = attr.ib(type=Node)


@attr.s(hash=True, repr=False)
class NotEqual(Node):
    lhs = attr.ib(type=Node)
    rhs = attr.ib(type=Node)


@attr.s(hash=True, repr=False)
class LessThan(Node):
    lhs = attr.ib(type=Node)
    rhs = attr.ib(type=Node)


@attr.s(hash=True, repr=False)
class LessThanEqual(Node):
    lhs = attr.ib(type=Node)
    rhs = attr.ib(type=Node)


@attr.s(hash=True, str=False)
class GreaterThan(Node):
    lhs = attr.ib(type=Node)
    rhs = attr.ib(type=Node)

    def __str__(self): 
        return f"{self.lhs.value} > {self.rhs.value}"


@attr.s(hash=True, repr=False)
class GreaterThanEqual(Node):
    lhs = attr.ib(type=Node)
    rhs = attr.ib(type=Node)


@attr.s(hash=True, repr=False)
class Relate(Node):
    vars = attr.ib(type=list) # list of vars that are imbued meaning as IV/DV/Variables at runtime during interpretation
    predictions = attr.ib(type=list, default=None) # list of Nodes

    @classmethod
    def from_str_names(cls, iv: str, dv: str, predictions: list):

        data = {'X': x, 'Y': y}
        df = pd.DataFrame.from_dict(data)

        x_var = Variable('X', dtype=DataType.INTERVAL, categories=None, drange=None)
        y_var = Variable('Y', dtype=DataType.INTERVAL, categories=None, drange=None)

        return cls(dfile='', variables=[x_var,y_var], data=df)


@attr.s(hash=True, repr=False)
class Mean(Node):
    var = attr.ib(type=Node)


# TODO: Should definitely check that the values that are passed are correct/exist/legit
@attr.s(hash=True, cmp=False)
class Literal(Node):
    value = attr.ib(type=Union[int, float, str])

    def __lt__(self, other):
        return LessThan(self, other)

    def __gt__(self, other):
        return GreaterThan(self, other)

    def __eq__(self, other):
        return Equal(self, other)
    
    def __ne__(self, other):
        return NotEqual(self, other)


@attr.s(hash=True, cmp=False)
class Relationship(Node):
    var = attr.ib(type=Variable)
    
    def positive(self, other):
        # import pdb; pdb.set_trace()
        # return Relate([self.var, other.var])
        return PositiveRelationship(self, other)

    def negative(self, other):
        return NegativeRelationship(self, other)


@attr.s(hash=True, cmp=False)
class PositiveRelationship(Node):
    lhs = attr.ib(type=Variable)
    rhs = attr.ib(type=Variable)


@attr.s(hash=True, cmp=False)
class NegativeRelationship(Node):
    lhs = attr.ib(type=Node)
    rhs = attr.ib(type=Node)


from tea.global_vals import *
from tea.helpers.constants.test_names import *
from enum import Enum
from .value import Value
from tea.ast import DataType, LessThan, GreaterThan, Literal, Relationship

# Other
import attr


# Need to be reset for each test
group1 = None
group2 = None
outcome = None

__test_to_statistic_of_interest__ = {
    STUDENTS_T_NAME: "mean",
    WELCHS_T_NAME: "mean",
    MANN_WHITNEY_NAME: "median",
    PAIRED_STUDENTS_NAME: "mean",
    WILCOXON_SIGNED_RANK_NAME: "median",
    KRUSKALL_WALLIS_NAME: "median",
    FACTORIAL_ANOVA_NAME: "mean",
}

__stats_tests_to_null_hypotheses__ = {
    PEARSON_NAME:  'There is no relationship between {} and {}.',
    KENDALLTAU_NAME: 'There is no relationship between {} and {}.',
    SPEARMAN_NAME: 'There is no monotonic relationship between {} and {}.',
    POINTBISERIAL_NAME: 'There is no association between {} and {}.',

    STUDENTS_T_NAME : 'There is no difference in {}s between {} and {} on {}.',
    WELCHS_T_NAME : 'There is no difference in {}s between {} and {} on {}.',
    MANN_WHITNEY_NAME : 'There is no difference in {}s between {} and {} on {}.',
    PAIRED_STUDENTS_NAME : 'There is no difference in {}s between {} and {} on {}.',
    WILCOXON_SIGNED_RANK_NAME : 'There is no difference in {}s between {} and {} on {}.',

    CHI_SQUARE_NAME : 'There is no association between {} and {} on {}.',
    FISHER_EXACT_NAME : 'There is no association between {} and {} on {}.',

    # Regression or anova?
    F_TEST_NAME : 'The variances of the groups ({}) are{}equal.',
    KRUSKALL_WALLIS_NAME : 'There is no difference in {}s between {} on {}.',
    # This isn't very descriptive…
    "Friedman" : f'There is no difference between the groups',
    FACTORIAL_ANOVA_NAME : 'There is no difference in {}s between {} on {}.',
    RM_ONE_WAY_ANOVA_NAME : 'The means of all groups/conditions ({}) are{}equal.',

    # Does this depend on the statistic being calculated?
    "Bootstrap" : f''
}

__two_group_tests__ = {
    PEARSON_NAME,
    KENDALLTAU_NAME,
    SPEARMAN_NAME,
    POINTBISERIAL_NAME,
}

__two_group_outcome_tests__ = {
    STUDENTS_T_NAME,
    WELCHS_T_NAME,
    MANN_WHITNEY_NAME,
    PAIRED_STUDENTS_NAME,
    WILCOXON_SIGNED_RANK_NAME,
}

__categorical_tests__ = {
    CHI_SQUARE_NAME,
    FISHER_EXACT_NAME,
}

__many_groups_test__ = {
    RM_ONE_WAY_ANOVA_NAME,
    F_TEST_NAME,
}

__many_groups_outcome_tests__ = {
    KRUSKALL_WALLIS_NAME,
    FACTORIAL_ANOVA_NAME,
}


@attr.s(init=True)
class TestResult(Value): 
    name = attr.ib()
    test_statistic = attr.ib()
    p_value = attr.ib()
    prediction = attr.ib()
    alpha = attr.ib()
    dof = attr.ib(default=None)
    adjusted_p_value = attr.ib(default=None)
    corrected_p_value = attr.ib(default=None) # for multiple comparisons
    null_hypothesis = attr.ib(default=None)
    interpretation = attr.ib(default=None)
    table = attr.ib(default=None)
    x = attr.ib(default=None)
    y = attr.ib(default=None)
    group_descriptive_statistics = attr.ib(default=None)

    def __attrs_post_init__(self):
        self.adjust_p_val()

        # TODO: Handle more cases without a prediction.
        if self.prediction or (self.x and self.y):
            self.null_hypothesis = self.get_null_hypothesis()
            self.set_interpretation()
        else:
            print("No prediction specified.")

    def adjust_p_val(self): 
        # Adjust p value if it has not already been adjusted
        if not self.adjusted_p_value:
            if self.prediction or (self.x and self.y):
                if self._is_one_sided():
                    self.adjusted_p_value = self.p_value/2
                else: 
                    self.adjusted_p_value = self.p_value

    def bonferroni_correction(self, num_comparisons):
        assert(num_comparisons >= 1)
        
        # Does it already have an adjusted p value?
        if self.adjusted_p_value:
            self.corrected_p_value = self.adjusted_p_value/(num_comparisons * 1.0)

        else: 
            if isinstance(self.p_value, float): # Self is not a result from Bootstrapping
                self.corrected_p_value = self.p_value/(num_comparisons * 1.0)

        ## FOR DEBUGGING
        # if num_comparisons != 1:
        #     import pdb; pdb.set_trace()

    def get_null_hypothesis(self):
        # TODO: Passing x and y seems more modular than passing string?
        if self.name in __two_group_tests__:
            var1, var2 = self._calculate_two_group_vars()
            return __stats_tests_to_null_hypotheses__[self.name] \
                .format(var1, var2)
        elif self.name in __two_group_outcome_tests__:
            assert self.x
            assert len(self.x.metadata[categories]) == 2
            assert self.y

            cats = list(self.x.metadata[categories].items())
            return __stats_tests_to_null_hypotheses__[self.name] \
                .format(__test_to_statistic_of_interest__[self.name], f"{self.x.metadata[name]} = {cats[0][0]}", f"{self.x.metadata[name]} = {cats[1][0]}", self.y.metadata[name])
        elif self.name in __many_groups_test__:
            assert self.x
            assert len(self.x.metadata[categories]) > 0

            return __stats_tests_to_null_hypotheses__[self.name] \
                .format(self._get_all_groups(), " ")
        elif self.name in __many_groups_outcome_tests__:
            assert self.x
            assert self.y

            return __stats_tests_to_null_hypotheses__[self.name] \
                .format(__test_to_statistic_of_interest__[self.name], self._get_all_groups(), self.y.metadata[name])
        elif self.name in __categorical_tests__:
            assert self.x
            assert self.y
            var1, var2 = self._calculate_two_group_vars()
            return __stats_tests_to_null_hypotheses__[self.name] \
                .format(var1, var2, self.y.metadata[name])
        return ""

    def set_interpretation(self):
        """
            Based on https://stats.idre.ucla.edu/other/mult-pkg/faq/general/faq-what-are-the-differences-between-one-tailed-and-two-tailed-tests/
            From the example on this site:
            If t is negative and we are checking a less than relationship, use p/2 as the adjusted p-value.
            If t is negative and we are checking a greater than relationship, use 1 - p/2 as the adjusted p-value.
        """

        # import pdb; pdb.set_trace()
        # one_sided = True if self.adjusted_p_value else False
        one_sided = self._is_one_sided()

        # Start interpretation
        ttest_result = Significance.not_significant
        if self.test_statistic > 0:
            if one_sided and isinstance(self.prediction, GreaterThan) and self.adjusted_p_value < self.alpha:
                ttest_result = Significance.significantly_greater
            elif one_sided and isinstance(self.prediction, LessThan) and 1 - self.adjusted_p_value < self.alpha:
                ttest_result = Significance.significantly_less
            elif not one_sided and self.adjusted_p_value and self.adjusted_p_value < self.alpha:
                ttest_result = Significance.significantly_different
            else:
                ttest_result = Significance.not_significant
        elif self.test_statistic < 0:
            if one_sided and isinstance(self.prediction, LessThan) and  1 - self.adjusted_p_value < self.alpha:
                ttest_result = Significance.significantly_less
            elif one_sided and isinstance(self.prediction, GreaterThan) and self.adjusted_p_value < self.alpha:
                ttest_result = Significance.significantly_greater
            elif not one_sided and self.adjusted_p_value < self.alpha:
                ttest_result = Significance.significantly_different
            else:
                ttest_result = Significance.not_significant
        elif not one_sided and self.adjusted_p_value and self.adjusted_p_value < self.alpha:
            ttest_result = ttest_result.significantly_different
        else:
            assert False, "test_statistic = 0 and it's not a one-sided test. Not sure under what conditions this is possible."


        # TODO Maybe the "means" part could be replacable....
        self.interpretation = f"t({self.dof}) = {'%.5f'%(self.test_statistic)}, p = {'%.5f'%(self.adjusted_p_value)}. " if self.dof else ""
        if ttest_result == ttest_result.not_significant:
            self.interpretation += f"Fail to reject the null hypothesis at alpha = {self.alpha}. "
            self.interpretation += self.null_hypothesis
            # self.interpretation = f"The difference in means of {y.metadata[name]} for {x.metadata[name]} = {self.prediction.lhs.value} " \
            #     f"and {x.metadata[name]} = {self.prediction.rhs.value} is not significant."
        elif ttest_result == ttest_result.significantly_different:
            self.interpretation += f"Reject the null hypothesis at alpha = {self.alpha}. "
            if self.name in __two_group_tests__:
                var1, var2 = self._calculate_two_group_vars()
                self.interpretation += f"There is a relationship between {var1} and {var2}."
            elif self.name in __two_group_outcome_tests__:
                var1, var2 = self._calculate_two_group_outcome_vars()
                stat = __test_to_statistic_of_interest__[self.name]
                self.interpretation += f"The difference in {stat}s of {self.y.metadata[name]} for {self.x.metadata[name]} = {var1} " \
                    f"and {self.x.metadata[name]} = {var2} is significant. "
            elif self.name in __many_groups_test__:
                self.interpretation += __stats_tests_to_null_hypotheses__[self.name] \
                    .format(self._get_all_groups(), " not ")
            elif self.name in __many_groups_outcome_tests__:
                stat = __test_to_statistic_of_interest__[self.name]
                self.interpretation += f"There is a difference in {stat}s of {self.y.metadata[name]} for at least one of {self._get_all_groups()}."
            elif self.name in __categorical_tests__:
                var1, var2 = self._calculate_two_group_vars()
                self.interpretation += f"There is an association between {var1} and {var2} on {self.y.metadata[name]}."

        elif ttest_result == ttest_result.significantly_greater:
            self.interpretation += f"Reject the null hypothesis at alpha = {self.alpha}. "
            stat = __test_to_statistic_of_interest__[self.name]
            var1, var2 = self._calculate_two_group_outcome_vars()
            self.interpretation += f"The {stat} of {self.y.metadata[name]} for {self.x.metadata[name]} = {var1} is significantly" \
                f" greater than the {stat} for {self.x.metadata[name]} = {var2}. "
        elif ttest_result == ttest_result.significantly_less:
            self.interpretation += f"Reject the null hypothesis at alpha = {self.alpha}. "
            stat = __test_to_statistic_of_interest__[self.name]
            var1, var2 = self._calculate_two_group_outcome_vars()
            self.interpretation += f"The {stat} of {self.y.metadata[name]} for {self.x.metadata[name]} = {var1} is significantly" \
                f" less than the {stat} for {self.x.metadata[name]} = {var2}. "
        else:
            assert False, "ttest_result case without an associated self.interpretation."

    # def adjust_p_val(self, correction): 
    #     self.self.adjusted_p_value = attr.ib()
    #     self.self.adjusted_p_value = self.p_value/correction
    
    # def set_adjusted_p_val(self, adjusted_p_value): 
    #     self.self.adjusted_p_value = attr.ib()
    #     self.self.adjusted_p_value = adjusted_p_value

    def add_effect_size(self, name, effect_size): 
        if hasattr(self, 'effect_size'):
            self.effect_size[name] = effect_size
        else: 
            self.effect_size = attr.ib()
            self.effect_size = {name: effect_size}

    def add_effect_size_to_interpretation(self):
        effect_sizes = ""
        for effect_size_name, effect_size_value in self.effect_size.items():
            effect_sizes += f"{effect_size_name} = {'%.5f' % effect_size_value}, "
        effect_sizes = effect_sizes[:-2]
        self.interpretation += f"The effect size is {effect_sizes}. " \
            f"The effect size is the magnitude of the difference, which gives a holistic view of the results [1].\n" \
            f"[1] Sullivan, G. M., & Feinn, R. (2012). Using effect size—or why the P value is not enough. Journal of graduate medical education, 4(3), 279-282."

    
    def add_doc(self, name, dof):
        import pdb; pdb.set_trace()

    def _calculate_two_group_vars(self):
        var1 = ""
        var2 = ""
        if self.prediction:
            if isinstance(self.prediction.lhs, Literal):
                var1 = self.prediction.lhs.value
                var2 = self.prediction.rhs.value
            else:
                assert isinstance(self.prediction.lhs, Relationship)
                # assert isinstance(self.prediction.lhs, Relationship)
                var1 = self.prediction.lhs.var
                var2 = self.prediction.rhs.var
        else:
            assert len(self.x.metadata[categories]) == 2
            cats = list(self.x.metadata[categories].items())
            var1 = cats[0][0]
            var2 = cats[1][0]

        return (var1, var2)

    def _calculate_two_group_outcome_vars(self):
        var1 = self.prediction.lhs.value
        var2 = self.prediction.rhs.value
        if self.group_descriptive_statistics:
            var1 = f"{var1} (M={'%.5f'%(self.group_descriptive_statistics[var1]['mean'])}, SD={'%.5f'%(self.group_descriptive_statistics[var1]['stdev'])})"
            var2 = f"{var2} (M={'%.5f'%(self.group_descriptive_statistics[var2]['mean'])}, SD={'%.5f'%(self.group_descriptive_statistics[var2]['stdev'])})"

        return (var1, var2)

    def _get_all_groups(self):
        assert self.x
        return self.x.metadata[name] + " = " + ', '.join([str(cat[0]) for cat in list(self.x.metadata[categories].items())])

    def _is_one_sided(self):
        return True if self.name in __two_group_outcome_tests__ and \
                        (isinstance(self.prediction, GreaterThan) or isinstance(self.prediction, LessThan)) \
                else False


class Significance(Enum): 
        not_significant = 0
        significantly_different = 1
        significantly_greater = 2
        significantly_less = 3


from tea.helpers.constants.default_values import DEFAULT_ALPHA_PARAMETER
from tea.global_vals import *
from tea.ast import *
import attr
from .value import Value

# CombinedData is the runtime data structure used to unify experimental design and variable declarations
@attr.s(init=True)
class CombinedData(Value):
    vars = attr.ib(default=[]) # list of VarData objects 
    study_type = attr.ib(default=observational_identifier)
    # set of characteristics about the groups that are used to determine statistical test
    alpha = attr.ib(type=float, default=DEFAULT_ALPHA_PARAMETER)
    properties = attr.ib(default=dict())
    

    def _update_vars(self):
        self.vars = self.get_explanatory_variables() + self.get_explained_variables()

    # Null Hypothesis: Groups come from populations with equal variance
    def has_equal_variance(self): 
        global eq_variance, alpha
        
        if self.properties[eq_variance]: 
            return self.properties[eq_variance][1] >= self.alpha  # Cannot reject null hypothesis of equal variances
        else: 
            return False

    def has_paired_observations(self):
        global paired 

        return self.properties[paired]

    def has_independent_observations(self):
        global paired
        
        return not self.properties[paired]

    def difference_between_paired_value_is_normal(self):
        if not self.has_paired_observations():
            return False
        assert False, "Implement this property to convey information about whether difference" \
                      "between paired values is normally distributed."
    
    # @return list of VarData instances that are in this object's 
    # vars that have the @param role
    def get_vars(self, role: str): 
        role_vars = []
        for v in self.vars:
            if v.role == role: 
                role_vars.append(v) 
        
        return role_vars

    def get_explanatory_variables(self): 
        explanatory_variables = None
        
        if self.study_type == experiment_identifier:
            explanatory_variables = self.get_vars(iv_identifier)
        else: 
            assert(self.study_type == observational_identifier)
            explanatory_variables  = self.get_vars(contributor_identifier)
            
        return explanatory_variables
    
    def get_explained_variables(self): 
        explained_variables = None
        
        if self.study_type == experiment_identifier:
            explained_variables = self.get_vars(dv_identifier)
        else: 
            assert(self.study_type == observational_identifier)
            explained_variables  = self.get_vars(outcome_identifier)
            
        return explained_variables

    def has_one_x(self): 
        xs = self.get_explanatory_variables()

        return len(xs) == 1
    
    def has_one_y(self): 
        ys = self.get_explained_variables()

        return len(ys) == 1


from tea.global_vals import *
from tea.helpers.constants.default_values import DEFAULT_ALPHA_PARAMETER
from tea.runtimeDataStructures.dataset import Dataset
from tea.runtimeDataStructures.varData import VarData
from tea.runtimeDataStructures.combinedData import CombinedData
from tea.runtimeDataStructures.bivariateData import BivariateData
from tea.helpers.evaluateHelperMethods import get_data, compute_normal_distribution, compute_eq_variance
# from tea.z3_solver.solver
from tea.logging.tea_logger import TeaLogger

import attr
import z3
from typing import Dict, List


# Prog -> List[StatisticalTest] -> Query
alpha = DEFAULT_ALPHA_PARAMETER
MODE = 'strict' # Default

# Contains a map from z3 variables representing tests
# back to the test objects allowing us to map back
# from a model back to the tests.
__test_map__ = {}

#  A list of all tests in the system which we can use
# to iterate over all tests.
__ALL_TESTS__ = None

def all_tests():
    """A helper for accessing the global set of tests"""
    global __ALL_TESTS__
    return __ALL_TESTS__

def reset_all_tests(): 
    global __test_map__, __ALL_TESTS__,__property_to_function__
    global pearson_corr, kendalltau_corr, pointbiserial_corr_a, pointbiserial_corr_b
    global students_t, paired_students_t, welchs_t, mannwhitney_u
    global chi_square, fishers_exact
    global f_test, rm_one_way_anova, kruskall_wallis, friedman, factorial_ANOVA

    pearson_corr = None
    kendalltau_corr = None
    spearman_corr = None
    pointbiserial_corr_a = None
    pointbiserial_corr_b = None
    students_t = None
    paired_students_t = None
    welchs_t = None 
    mannwhitney_u = None
    wilcoxon_signed_rank = None
    chi_square = None
    fishers_exact = None
    f_test = None    
    rm_one_way_anova = None
    kruskall_wallis = None
    friedman = None
    factorial_ANOVA = None
    
    __test_map__ = {}
    __ALL_TESTS__ = None
    __property_to_function__ = {}
    __property_var_map__ = {}

# Contains the global map from z3 variables which
# represent properties applied to variables back
# the properties that were applied.
__property_var_map__ = {}

# Contains a map from z3 variables representing the props
# back to the Property objects
__property_map__ = {}

# List of all Properties in the system which we can use 
# to iterate over all properties 
__ALL_PROPERTIES__ = []

# Map of all properties to functions that
# will give values to the properties
__property_to_function__ = {}


def all_props(): 
    """A helper for accessing the global set of properties"""
    global __ALL_PROPERTIES__
    return __ALL_PROPERTIES__

# def subset_props(scope): 
#     """A helper for accessing subset of properties based on their scope (Test or variable, for now)"""
#     props = all_props()
#     selected_props = []
#     for p in props: 
#         if p.scope == scope:
#             selected_props.append(p)

#     return selected_props


def set_mode(mode): 
    global MODE
    MODE = mode


@attr.s(hash=False, cmp=False, auto_attribs=True, init=False)
class StatVar:
    name: str

    def __init__(self, name):
        self.name = name
        self.__z3__ = z3.Bool(self.name)


# TODO write test that these areq unique with pointer eq
# i.e StatVar('x') != StatVar('x')
# vs
# x = StatVar('x')
# assert x == x

class StatisticalTest:
    name: str
    variables: List[StatVar]
    test_properties: List["Property"]

    # properties_for_vars maps from a property to lists of indices
    # into test_vars that the property applies to. If the property
    # has arity 1, then it maps to a list of lists of length 1.
    # Similarly, if the property has arity 2, then it maps to
    # a list of lists of length 2.
    properties_for_vars: Dict["Property", List[List[int]]]

    # This needs to store number of variables
    # each property needs to be mapped to variable
    # 
    def __init__(self, name, test_vars, test_properties=None, properties_for_vars=None):
        global __test_map__, __ALL_TESTS__
        self.name = name
        self.test_vars = test_vars

        if test_properties is None:
            test_properties = []

        if properties_for_vars is None:
            properties_for_vars = {}

        self.test_properties = test_properties

        self.properties_for_vars = {}
        for key in properties_for_vars:
            # properties_for_vars is a list of lists of variables. Get the indices of the variables
            # in each list.
            for prop_args in properties_for_vars[key]:
                indices = []
                # if self.name == 'pointbiserial_corr_a': 
                #     import pdb; pdb.set_trace()
                # if self.name == 'factorial_ANOVA': 
                #     import pdb; pdb.set_trace()
                indices = [self.test_vars.index(arg) for arg in prop_args]
                if key not in self.properties_for_vars.keys():
                    self.properties_for_vars[key] = []
                self.properties_for_vars[key].append(indices)

        # Create variable.
        self.__z3__ = z3.Bool(self.name)

        # Populate global table.
        __test_map__[self.__z3__] = self
        if not __ALL_TESTS__: 
            __ALL_TESTS__ = []
        __ALL_TESTS__.append(self)

    # Question: does this mean you can't consider two tests of the same type on different variables at once?
    def apply(self, *test_vars):
        # if len(test_vars) != len(self.test_vars): 
        #     import pdb; pdb.set_trace()
        assert len(test_vars) == len(self.test_vars)
        self.test_vars = test_vars

    def _populate_properties(self):
        # Populate all props
        self._properties = []
        for prop in self.test_properties:
            # This is weird, and doesn't really fit into property model
            # Alternative: Assign test-level properties to all the variables rather than
            # the test name
            # self._properties.append(prop(StatVar(self.name)))
            variables = [v for v in self.test_vars]
            self._properties.append(prop(*variables))

        for prop in self.properties_for_vars:
            list_of_variable_indices = self.properties_for_vars[prop]
            for variable_indices in list_of_variable_indices:
                variable_indices = [self.test_vars[i] for i in variable_indices]
                self._properties.append(prop(*variable_indices))
        
    def properties(self):
        return self.test_properties + list(self.properties_for_vars)

    def query(self):  # May want to change this....
        global __property_to_function__

        self._populate_properties()  # Apply to specific instance of variables
        # if self.__query__ is None:
        conj = []
        for p in self._properties:  # combines Test and Var-specific properties
            # p.__z3__ = uninterpreted function
            # p.__var__ = instantiated z3 BoolVal
            # Add the uf, and interpret the uf as always ==ing the instantiated BoolVal
            # conj += [p.__z3__, p.__z3__ == p.__var__]
            conj += [p.__z3__]

            # Add property to property to function map
            __property_to_function__[p.__z3__] = p.property.function

        return conj
        # self.__query__ = query
        # return self.__query__

    @staticmethod
    def get_by_z3_var(var):
        """
        This relationship holds:
            t1 == StatisticalTet.get_by_z3_var(t1.__var__)
        """
        global __test_map__
        return __test_map__.get(var)


class Property:
    name: str
    description: str
    arity: int

    def __init__(self, name, description, function=None, arity=1):
        global __property_map__, __ALL_PROPERTIES__

        self.name = name
        self.description = description
        self.function = function
        self.arity = arity
        args = []
        for _ in range(self.arity):
            args.append(z3.BoolSort())
        args.append(z3.BoolSort())
        self.__z3__ = z3.Function(self.name, *args)  # e.g. continuous(x)
        self.__cache__ = {}

        # Populate global table.    
        __property_map__[self.__z3__] = self
        __ALL_PROPERTIES__.append(self)

    def __str__(self):
        return f"property:{self.name}"

    def __repr__(self):
        return f"Property({self.name})"

    def __eq__(self, other):
        return self.arity == other.arity and self.name == other.name and self.function == other.function

    def __hash__(self):
        return hash((self.arity, self.name, self.function))

    def __call__(self, *var_names):
        # import pdb; pdb.set_trace()
        if len(var_names) != self.arity:
            raise Exception(f"{self.name} property has arity {self.arity} " \
                            f"found {len(var_names)} arguments")
        cached = self.__cache__.get(tuple(var_names))
        if cached:
            return cached
        
        ap = AppliedProperty(self, var_names)
        # self.__cache__[tuple(var_names)] = ap
        return ap

    # Only used when the property must be changed depending on input data
    def _update(self, arity):
        self.arity = arity
        self.reset()

    def reset(self):
        args = []
        for _ in range(self.arity):
            args.append(z3.BoolSort())
        args.append(z3.BoolSort())
        self.__z3__ = z3.Function(self.name, *args)  # e.g. continuous(x)


class AppliedProperty:
    property: Property

    def __init__(self, prop, pvars):
        global __property_var_map__
        self.property = prop  # e.g. continuous
        self.vars = pvars  # (StatVar(name='x'),)
        # self._name = ""
        z3_args = []
        for tv in pvars:
            # Allows for unique identification of prop -> var, but not looking up from model because that refs prop name
            # self._name += tv.name + ":"
            # self._name + tv.name
            # import pdb; pdb.set_trace()
            z3_args.append(tv.__z3__)
        self._name = self.property.name  # continuous
        # Why is property fn a bool? Does this allow continuous(x) and not continuous(y)?
        # Answer: I think the undefined function is able to capture all functionality of
        # deciding the boolean property value for each variable, so the z3.Bool() for the
        # property name isn't necessary.
        # self.__var__ = z3.Bool(self._name)  # e.g. continuous
        self.__z3__ = prop.__z3__(*z3_args)  # e.g. continuous(x)

        # _name needs to be unique to avoid overwriting same property for different variables.
        # But if it is unique, it makes it more difficult to look up from the z3 model.
        # Solution: Make the name the same, but have it map to a list of variables that the
        # property applies to.
        if self._name not in __property_var_map__.keys():
            __property_var_map__[self._name] = []

        # TODO: When does __property_var_map__ need to be cleared?
        __property_var_map__[self._name].append(self.vars)

        self.property_test_results = None

    def __str__(self):
        return f"property_for_var:{self._name}"

    def __repr__(self):
        return f"AppliedProperty({self._name})"
    
    def __eq__(self, other):
        return self.property == other.property and self.vars == other.vars

    @staticmethod
    def get_by_z3_var(name):
        global __property_var_map__
        return __property_var_map__.get(name)


# Functions to verify properties
def is_bivariate(dataset: Dataset, var_data: CombinedData, alpha):
    return len(var_data.vars) == 2
    # could also do...
    # return isinstance(var_data, BivariateData)


def is_multivariate(datset: Dataset, var_data: CombinedData, alpha):
    return len(var_data.vars) > 2
    # could also do...
    # return isinstance(var_data, MultivariateData)


def has_one_x(dataset: Dataset, var_data: CombinedData, alpha): 
    xs = var_data.get_explanatory_variables()

    return len(xs) == 1


def has_one_y(dataset: Dataset, var_data: CombinedData, alpha): 
    ys = var_data.get_explained_variables()

    return len(ys) == 1


def has_paired_observations(dataset: Dataset, var_data: CombinedData, alpha):
    global paired

    return var_data.properties[paired]


def has_independent_observations(dataset: Dataset, var_data: CombinedData, alpha):
    global paired

    return not var_data.properties[paired]


def greater_than_5_frequency(dataset: Dataset, var_data: CombinedData, alpha): 
    xs = var_data.get_explanatory_variables()
    ys = var_data.get_explained_variables()

    if len(xs) == 1: 
        if len(ys) == 1: 
            x = xs[0]
            y = ys[0]

            if x.is_categorical() and y.is_categorical(): 

                # Get the count for each category
                x_cat = [k for k , v in x.metadata[categories].items()]
                y_cat = [k for k , v in y.metadata[categories].items()]

                for xc in x_cat: 
                    for yc in y_cat: 
                        data = dataset.select(y.metadata[name], where=[f"{x.metadata[name]} == '{xc}'", f"{y.metadata[name]} == '{yc}'"])                    

                        # Check that the count is at least five for each of the (x,y) group pairs
                        if len(data) < 5:
                            return False
                
                return True
            else: 
                return False
        else: 
            raise ValueError(f"Currently, chi square requires/only supports 1 explained variable, instead received: {len(ys)} -- {ys}")    
    else: 
        x0 = xs[0]
        x1 = xs[1]
        
        if x0.is_categorical() and x1.is_categorical():
            # Get the count for each category
            x0_cat = [k for k , v in x0.metadata[categories].items()]
            x1_cat = [k for k , v in x1.metadata[categories].items()]

            for x0c in x0_cat: 
                for x1c in x1_cat: 
                    data = dataset.select(x1.metadata[name], where=[f"{x.metadata[name]} == '{xc}'", f"{x1.metadata[name]} == '{x1c}'"])                    

                    # Check that the count is at least five for each of the (x,x1) group pairs
                    if len(data) < 5:
                        return False
            return True
        else: 
            return False



def has_equal_variance(dataset: Dataset, var_data: list, alpha):
    xs = []
    ys = []
    cat_xs = []
    cont_ys = []
    grouped_data = []


    if isinstance(var_data, CombinedData):
        xs = var_data.get_explanatory_variables()
        ys = var_data.get_explained_variables()

    else: 
        for var in var_data: 
            if var.role == iv_identifier or var.role == contributor_identifier:
                xs.append(var)
            if var.role == dv_identifier or var.role == outcome_identifier:
                ys.append(var)

    for x in xs: 
        if x.is_categorical(): 
            cat_xs.append(x)
    
    for y in ys: 
        if y.is_continuous(): 
            cont_ys.append(y)
    
    eq_var = (None, None)
    if cat_xs and cont_ys: 
        for y in ys:
            for x in xs: 
                cat = [k for k,v in x.metadata[categories].items()]
                for c in cat: 
                    data = dataset.select(y.metadata[name], where=[f"{x.metadata[name]} == '{c}'"])
                    grouped_data.append(data)
                if isinstance(var_data, BivariateData):
                    # Equal variance
                    eq_var = compute_eq_variance(grouped_data)
                else: 
                    eq_var = compute_eq_variance(grouped_data)

    if eq_var[0] is None and eq_var[1] is None:
        # import pdb; pdb.set_trace()
        # raise Exception("did not compute variance, this is a bug")
        return False

    return eq_var[1] > alpha


def has_groups_normal_distribution(dataset, var_data, alpha):
    xs = []
    ys = []
    cat_xs = []
    cont_ys = []
    grouped_data = []
    result = None

    if isinstance(var_data, CombinedData):
        xs = var_data.get_explanatory_variables()
        ys = var_data.get_explained_variables()

    else: 
        for var in var_data: 
            if var.role == iv_identifier or var.role == contributor_identifier:
                xs.append(var)
            if var.role == dv_identifier or var.role == outcome_identifier:
                ys.append(var)

    for x in xs: 
        if x.is_categorical(): 
            cat_xs.append(x)
    
    for y in ys: 
        if y.is_continuous(): 
            cont_ys.append(y)
    
    if cat_xs and cont_ys: 
        for y in ys:
            for x in xs: 
                cat = [k for k,v in x.metadata[categories].items()]
                for c in cat: 
                    if isinstance(c, str):
                        data = dataset.select(y.metadata[name], where=[f"{x.metadata[name]} == '{c}'"])
                    else: 
                        data = dataset.select(y.metadata[name], where=[f"{x.metadata[name]} == {c}"])
                    grouped_data.append(data)

                for group in grouped_data:
                    result = compute_normal_distribution(group)
                    if result[1] <= alpha:
                        return False, result

    return True, result


def is_categorical_var(dataset, var_data, alpha):
    assert(len(var_data) == 1)
    assert(isinstance(var_data[0], VarData))

    return var_data[0].is_categorical()


def has_two_categories(dataset, var_data, alpha): 
    assert(len(var_data) == 1)
    assert(isinstance(var_data[0], VarData))

    # First check that the variable is categorical
    assert(is_categorical_var(dataset, var_data, alpha))
    return len(var_data[0].metadata[categories].keys()) == 2


def has_two_or_more_categories(dataset, var_data, alpha): 
    assert(len(var_data) == 1)
    assert(isinstance(var_data[0], VarData))

    # First check that the variable is categorical
    assert(is_categorical_var(dataset, var_data, alpha))
    return len(var_data[0].metadata[categories].keys()) >= 2


def has_three_or_more_categories(dataset, var_data, alpha): 
    assert(len(var_data) == 1)
    assert(isinstance(var_data[0], VarData))

    # First check that the variable is categorical
    assert(is_categorical_var(dataset, var_data, alpha))
    return len(var_data[0].metadata[categories].keys()) >= 3


def is_continuous_var(dataset, var_data, alpha):
    assert(len(var_data) == 1)
    assert(isinstance(var_data[0], VarData))

    return var_data[0].is_continuous()


def is_continuous_or_ordinal_var(dataset, var_data, alpha):
    assert(len(var_data) == 1)
    assert(isinstance(var_data[0], VarData))

    return is_continuous_var(dataset, var_data, alpha) or var_data[0].is_ordinal()


def has_normal_distribution(dataset, var_data, alpha):
    assert(len(var_data) == 1)
    assert(isinstance(var_data[0], VarData))

    # Must be continuous to be normally distributed
    assert(is_continuous_var(dataset, var_data, alpha))
    # Get data from datasest using var_data's query 
    data = get_data(dataset, var_data[0])
    norm_test_results = compute_normal_distribution(data)

    return (norm_test_results[1] > alpha), norm_test_results


# Test properties
bivariate = Property('is_bivariate', "Exactly two variables involved in analysis", is_bivariate)
multivariate = Property('is_multivariate', "More than two variables involved in analysis", is_multivariate) # May not need this!
one_x_variable = Property('has_one_x', "Exactly one explanatory variable", has_one_x)
one_y_variable = Property('has_one_y', "Exactly one explained variable", has_one_y)
paired_obs = Property('has_paired_observations', "Paired observations", has_paired_observations)
independent_obs = Property('has_independent_observations', "Independent (not paired) observations", has_independent_observations)
greater_than_5_freq = Property('greater_than_5_freq', "Has a large sample size", greater_than_5_frequency, arity=2)

test_props = [bivariate, one_x_variable, one_y_variable, paired_obs, independent_obs]

# Variable properties
categorical = Property('is_categorical', "Variable is categorical", is_categorical_var)
two_categories = Property('has_two_categories', "Variable has two categories", has_two_categories)
two_or_more_categories = Property('has_two_or_more_categories', "Variable has two or more categories", has_two_or_more_categories)
three_or_more_categories = Property('has_three_or_more_categories', "Variable has three or more categories", has_three_or_more_categories)
# all_x_variables_categorical = Property('has_all_x_categorical', "All explanatory variables are categorical", 'variable')
# two_x_variable_categories = Property('has_two_categories_x_var', "Exactly two categories in explanatory variable", 'variable')
continuous = Property('is_continuous', "Continuous (not categorical) data", is_continuous_var)
# We could create a disjunction of continuous \/ ordinal instead
continuous_or_ordinal = Property('is_continuous_or_ordinal', "Continuous OR ORDINAL (not nominal) data", is_continuous_or_ordinal_var)
groups_normal = Property('is_groups_normal', "Groups are normally distributed", has_groups_normal_distribution, arity=2)
normal = Property('is_normal', "Normal distribution", has_normal_distribution)
eq_variance = Property('has_equal_variance', "Equal variance", has_equal_variance, arity=2)


def construct_test_axioms(solver):
    solver.add(paired_obs.__z3__ != independent_obs.__z3__)
    

def construct_axioms(variables):  # List[StatVar]
    _axioms = []
    for var in variables:
        # A variable must be continuous or categorical, but not both.
        _axioms.append(z3.And(z3.Or(continuous(var).__z3__, categorical(var).__z3__),
                              z3.Not(z3.And(continuous(var).__z3__, categorical(var).__z3__))))

        # If a variable is an explanatory variable and all explanatory variables are categorical,
        # then the variable must be categorical.
        # It isn't clear how to reason about whether a variable is an explanatory or explained variable.
        # _axioms.append(z3.Implies(all_x_variables_categorical(var).__z3__, categorical(var).__z3__))

        # Not sure how to reason about test properties like one_x_variable and one_y_variable.
        # _axioms.append(z3.Not(z3.And(one_x_variable(var).__z3__, one_y_variable(var).__z3__)))

        # If a variable is normal, then it cannot be categorical.
        _axioms.append(z3.Implies(normal(var).__z3__, z3.Not(categorical(var).__z3__)))

        # If a variable is continuous or ordinal, it must be continuous.
        _axioms.append(z3.Implies(continuous_or_ordinal(var).__z3__, continuous(var).__z3__))

        # If a variable has two categories, then it must be categorical.
        # _axioms.append(z3.Implies(two_x_variable_categories(var).__z3__, categorical(var).__z3__))

    return _axioms


bivariate_constructed = False
multivariate_constructed = False


# Constructs all the tests once
# TODO: Move to a separate file/location
def construct_all_tests(combined_data: CombinedData): 
    # global bivariate_constructed, multivariate_constructed

    # if not bivariate_constructed: 
    #     construct_bivariate_tests(combined_data)
    #     bivariate_constructed = True

    # if not multivariate_constructed: 
    #     construct_mutlivariate_tests(combined_data)
    #     multivariate_constructed = True

    construct_bivariate_tests(combined_data)
    construct_mutlivariate_tests(combined_data)


# The generic StatisticalTests have specific predefined arity.
# Some multivariate statistical tests, however, have various arities. 
# Therefore, support the construction of muliple generic StatisticalTests. 
def construct_mutlivariate_tests(combined_data):
    construct_factorial_ANOVA(combined_data)
    # pass


factorial_ANOVA = None


def construct_factorial_ANOVA(combined_data: CombinedData): 
    global factorial_ANOVA

    num_vars = len(combined_data.vars)

    x_vars = []
    y_vars = []
    all_vars = []
    for i in range(num_vars): 
        name = 'x' + str(i)
        # combined_data.vars[i].get_name()
        # import pdb; pdb.set_trace()
        if (i <= num_vars - 2): # All but the last var
            x = StatVar(name)
            x_vars.append(x)
            all_vars.append(x)
        else:
            assert(i == num_vars - 1)
            name = 'y' + str(i)
            y = StatVar(name)
            y_vars.append(y) # last var
            all_vars.append(y)
    assert(num_vars == len(x_vars) + len(y_vars))

    list_x_vars = [[v] for v in x_vars]
    # list_y_vars = [[v] for v in y_vars]
    # list_all_vars = list_x_vars + list_y_vars

    assert(len(y_vars) == 1)
    pairs_list = [[x, y_vars[0]] for x in x_vars]

    factorial_ANOVA = StatisticalTest('factorial_ANOVA', all_vars, # Variable number of factors
                                test_properties=
                                [one_y_variable],
                                # [one_y_variable, two_or_more_x_variables],
                                properties_for_vars={
                                continuous: [y_vars],
                                categorical: list_x_vars, # Variable number of factors
                                two_or_more_categories: list_x_vars,
                                groups_normal: pairs_list, # Variable number of factors
                                eq_variance: pairs_list # Variable number of factors
                                }) 


pearson_corr = None
kendalltau_corr = None
spearman_corr = None
pointbiserial_corr_a = None
pointbiserial_corr_b = None
students_t = None
paired_students_t = None
welchs_t = None 
mannwhitney_u = None
chi_square = None
fishers_exact = None 
f_test = None    
rm_one_way_anova = None
kruskall_wallis = None
friedman = None


def construct_bivariate_tests(combined_data: CombinedData): 
    global pearson_corr, kendalltau_corr, spearman_corr, pointbiserial_corr_a, pointbiserial_corr_b
    global students_t, paired_students_t, welchs_t, mannwhitney_u
    global chi_square, fishers_exact
    global f_test, kruskall_wallis, friedman, rm_one_way_anova

    # Bivariate analyses only make sense when there are two variables
    # in combined_data
    if len(combined_data.vars) == 2: 

        # CORRELATIONS
        x0 = StatVar('x0')
        x1 = StatVar('x1')

        pearson_corr = StatisticalTest('pearson_corr', [x0, x1],
                                        test_properties=
                                        [bivariate],
                                        properties_for_vars={
                                            continuous: [[x0], [x1]],
                                            normal: [[x0], [x1]]
                                        })

        kendalltau_corr = StatisticalTest('kendalltau_corr', [x0, x1],
                                        test_properties=
                                        [bivariate],
                                        properties_for_vars={
                                            continuous_or_ordinal: [[x0], [x1]]
                                        })

        spearman_corr = StatisticalTest('spearman_corr', [x0, x1],
                                        test_properties=
                                        [bivariate],
                                        properties_for_vars={
                                            continuous_or_ordinal: [[x0], [x1]]
                                        })

        # Need both? in case order of categortical and continuous differs?
        # TODO: Could just sort before apply test?
        pointbiserial_corr_a = StatisticalTest('pointbiserial_corr_a', [x0, x1],
                                        test_properties=
                                        [bivariate, independent_obs],
                                        properties_for_vars={
                                            continuous: [[x1]],
                                            # normal: [[x1]],
                                            groups_normal: [[x0, x1]],
                                            categorical: [[x0]],
                                            two_categories: [[x0]],
                                            eq_variance: [[x0, x1]]

                                        })

        pointbiserial_corr_b = StatisticalTest('pointbiserial_corr_b', [x0, x1],
                                        test_properties=
                                        [bivariate, independent_obs],
                                        properties_for_vars={
                                            continuous: [[x0]],
                                            # normal: [[x0]],
                                            groups_normal: [[x1, x0]],
                                            categorical: [[x1]],
                                            two_categories: [[x1]],
                                            eq_variance: [[x1, x0]]
                                        })                                

        # T-TESTS
        x = StatVar('x')
        y = StatVar('y')

        students_t = StatisticalTest('students_t', [x, y],
                                    test_properties=
                                        [bivariate, one_x_variable, one_y_variable, independent_obs],
                                    properties_for_vars={
                                        categorical : [[x]],
                                        two_categories: [[x]],
                                        continuous: [[y]],
                                        eq_variance: [[x, y]],
                                        groups_normal: [[x, y]]
                                        })

        welchs_t = StatisticalTest('welchs_t', [x, y],
                                    test_properties=
                                        [bivariate, one_x_variable, one_y_variable, independent_obs],
                                    properties_for_vars={
                                        categorical : [[x]],
                                        two_categories: [[x]],
                                        continuous: [[y]],
                                        groups_normal: [[x, y]]
                                        #   groups_normal: [[y]], # TODO: Check that each group is normally distributed
                                        })

        mannwhitney_u = StatisticalTest('mannwhitney_u', [x, y],
                                    test_properties=
                                        [one_x_variable, one_y_variable, independent_obs],
                                    properties_for_vars={
                                        categorical : [[x]],
                                        two_categories: [[x]],
                                        continuous_or_ordinal: [[y]],
                                        # conflicting sources, but remove for now
                                        #   eq_variance: [x, y] # Is this an assumption of mann whitney??
                                        })       
        
        paired_students_t = StatisticalTest('paired_students_t', [x, y],
                                    test_properties=
                                        [bivariate, one_x_variable, one_y_variable, paired_obs],
                                    properties_for_vars={
                                        categorical : [[x]],
                                        two_categories: [[x]],
                                        continuous: [[y]],
                                        groups_normal: [[x, y]],
                                        eq_variance: [[x, y]]
                                        })  
        
        wilcoxon_signed_rank = StatisticalTest('wilcoxon_signed_rank', [x, y],
                                    test_properties=
                                        [bivariate, one_x_variable, one_y_variable, paired_obs],
                                    properties_for_vars={
                                        categorical : [[x]],
                                        two_categories: [[x]],
                                        continuous: [[y]],
                                        })

        # CONTINGENCY TABLES (Categorical data)
        chi_square = StatisticalTest('chi_square', [x, y],
                                        test_properties=
                                        [bivariate, independent_obs, greater_than_5_freq],
                                        properties_for_vars={
                                            categorical:  [[x], [y]],
                                            two_or_more_categories: [[x],[y]]
                                        })                       

        fishers_exact = StatisticalTest('fishers_exact', [x, y],
                                        test_properties=
                                        [bivariate, independent_obs],
                                        properties_for_vars={
                                            categorical:  [[x], [y]],
                                            two_categories: [[x],[y]]
                                        })

        # ANOVAs
        f_test = StatisticalTest('f_test', [x, y], # Variable number of factors
                                    test_properties=
                                    [independent_obs, one_x_variable, one_y_variable],
                                    properties_for_vars={
                                    continuous: [[y]],
                                    categorical: [[x]], # Variable number of factors
                                    two_or_more_categories: [[x]],
                                    groups_normal: [[x, y]], # Variable number of factors
                                    eq_variance: [[x, y]] # Variable number of factors
                                    }) 
        
        kruskall_wallis = StatisticalTest('kruskall_wallis', [x, y], # Variable number of factors
                                    test_properties=
                                    [independent_obs, one_x_variable, one_y_variable],
                                    properties_for_vars={
                                    continuous: [[y]],
                                    categorical: [[x]], # Variable number of factors
                                    two_or_more_categories: [[x]]
                                    })

        
        rm_one_way_anova = StatisticalTest('rm_one_way_anova', [x, y], # Variable number of factors
                                test_properties=
                                [paired_obs, one_x_variable, one_y_variable],
                                properties_for_vars={
                                continuous: [[y]],
                                categorical: [[x]], # Variable number of factors
                                two_or_more_categories: [[x]],
                                groups_normal: [[x, y]], # Variable number of factors
                                eq_variance: [[x, y]] # Variable number of factors
                                }) 

        friedman = StatisticalTest('friedman', [x,y], # Variable number of factors
                                test_properties=
                                [paired_obs, one_x_variable, one_y_variable],
                                properties_for_vars={
                                continuous: [[y]],
                                categorical: [[x]], # Variable number of factors
                                three_or_more_categories: [[x]],
                                }) 


"""

# Specify variable properties as list of lists so that the same property can
# apply to multiple variables individually. If "categorical" is specified twice,
# only the second variable is kept because it overwrites the first entry in the
# dictionary.
chi_square_test = StatisticalTest('chi_square', [x, y],
                                   test_properties=[one_x_variable, one_y_variable],
                                   properties_for_vars={
                                       categorical: [[x], [y]]
                                   })
"""


# Verify the property against data
def verify_prop(dataset: Dataset, combined_data: CombinedData, prop:AppliedProperty):
    global alpha

    if len(prop.vars) == len(combined_data.vars):
        kwargs = {'dataset': dataset, 'var_data': combined_data, 'alpha': alpha}
        # import pdb; pdb.set_trace()
        if __property_to_function__ == {}:
            # import pdb; pdb.set_trace()
            prop_val = prop.property.function(**kwargs)
        else: 
            prop_val = __property_to_function__[prop.__z3__](**kwargs)   
    else: 
        assert (len(prop.vars) < len(combined_data.vars))
        var_data = []
        # For each of the variables for which we are checking the current prop
        for test_var in prop.vars:
            # For each variable in all variables in combined data
            for var in combined_data.vars:
                if var.metadata[name] == test_var.name:
                    var_data.append(var)
        kwargs = {'dataset': dataset, 'var_data': var_data, 'alpha': alpha}
        if __property_to_function__ == {}:
            prop_val = prop.property.function(**kwargs)
        else: 
            prop_val = __property_to_function__[prop.__z3__](**kwargs)

    ret_val = None
    if isinstance(prop_val, tuple):
        assert len(prop_val) == 2
        ret_val = prop_val[0]
        prop.property_test_results = prop_val[1]
    else:
        ret_val = prop_val
    
    return ret_val


# Assumes properties to hold
def assume_properties(stat_var_map, assumptions: Dict[str,str], solver, dataset, combined_data): 
    global assumptions_to_properties, alpha
    
    tea_logger = TeaLogger.get_logger()

    assumed_props = []

    # Go through all assumptions
    for a in assumptions: 
        if a in alpha_keywords:
            alpha = float(assumptions[a])
        elif a in assumptions_to_properties:
            # Apply corresponding properties to all variables for which the property is assumed
            for prop in all_props(): 
                if prop.name in assumptions_to_properties[a]: 
                    # Convert assumptions to use correct statistical variable names.
                    stat_vars = []
                    for var in assumptions[a]:
                        # This is really hacky -- should probably change the 
                        # incoming assumptions data structure before calling this method
                        if isinstance(var, str):
                            assert var in stat_var_map 
                            # stat_vars.append(stat_var_map[var]) 
                            ap = prop(stat_var_map[var]) 
                            assumed_props.append(ap)

                            # CHECK ASSUMPTIONS HERE
                            val = verify_prop(dataset, combined_data, ap)
                            if MODE == 'strict': 
                                tea_logger.log_debug(f"Running under STRICT mode.")
                                if val: 
                                    tea_logger.log_debug(f"User asserted property: {prop.name} is supported by statistical checking. Tea agrees with the user.")
                                else: 
                                    tea_logger.log_debug(f"User asserted property: {prop.name}, but is NOT supported by statistical checking. Tea will override user assertion.") 
                                solver.add(ap.__z3__ == z3.BoolVal(val))
                            elif MODE == 'relaxed': 
                                tea_logger.log_debug(f"Running under RELAXED mode.")
                                if val: 
                                    tea_logger.log_debug(f"User asserted property: {prop.name} is supported by statistical checking. Tea agrees with the user.")
                                else: 
                                    tea_logger.log_debug(f"User asserted property: {prop.name}, but is NOT supported by statistical checking. User assertion will be considered true.")
                                
                                solver.add(ap.__z3__ == z3.BoolVal(True))
                            else: 
                                raise ValueError(f"Invalid MODE: {MODE}")
                        else: 
                            assert isinstance(var, list)
                            stat_vars = [stat_var_map[v] for v in var]
                            ap = prop(*stat_vars)
                            assumed_props.append(ap)

                            # CHECK ASSUMPTIONS HERE
                            val = verify_prop(dataset, combined_data, ap)
                            if MODE == 'strict': 
                                tea_logger.log_debug(f"Running under STRICT mode.")
                                if val: 
                                    tea_logger.log_debug(f"User asserted property: {prop.name} is supported by statistical checking. Tea agrees with the user.")
                                else: 
                                    tea_logger.log_debug(f"User asserted property: {prop.name}, but is NOT supported by statistical checking. Tea will override user assertion.")
                                solver.add(ap.__z3__ == z3.BoolVal(val))
                            elif MODE == 'relaxed': 
                                tea_logger.log_debug(f"Running under RELAXED mode.")
                                if val: 
                                    tea_logger.log_debug(f"User asserted property: {prop.name} is supported by statistical checking. Tea agrees with the user.")
                                else: 
                                    tea_logger.log_debug(f"User asserted property: {prop.name}, but is NOT supported by statistical checking. User assertion will be considered true.")
                                solver.add(ap.__z3__ == z3.BoolVal(True)) # override user
                            else: 
                                raise ValueError(f"Invalid MODE: {MODE}")
        else:
            pass
    return assumed_props


# Helper for synthesize tests
def is_assumed_prop(assumed_props, prop):
    for ap in assumed_props:
        if prop == ap: 
            return True
    
    return False


# Problem statement: Given a set of properties, tell me which tests are valid to run
# This is a concrete (rather than symbolic) problem 
# @param combined_data CombinedData object
# @returns list of Property objects that combined_data exhibits
def synthesize_tests(dataset: Dataset, assumptions: Dict[str,str], combined_data: CombinedData):
    reset_all_tests()
    
    tea_logger = TeaLogger.get_logger()
    construct_all_tests(combined_data)

    global name
    stat_var_map = {}

    # Reorder variables so that y var is at the end
    combined_data._update_vars() 

    # Compute unique statisical variable names from the combined data.
    combined_data_vars = []
    for v in combined_data.vars:
        var = StatVar(v.metadata[name])
        stat_var_map[v.metadata[name]] = var 
        combined_data_vars.append(var)

    # Assume properties based on user assumptions and mode
    solver = z3.Solver()
    # s = Tactic('qflia').solver()
    assumed_props = assume_properties(stat_var_map, assumptions, solver, dataset, combined_data)
    # import pdb; pdb.set_trace()

    # Update the arity of test-level properties
    for prop in test_props: 
        prop._update(len(combined_data.vars))

    # Apply all tests to the variables we are considering now in combined_data
    for test in all_tests(): 
        test.apply(*combined_data_vars)

    solver.push() # Create backtracking point
    model = None # Store model

    # For each test, add it to the solver as a constraint. 
    # Add the tests and their properties
    for test in all_tests():
        tea_logger.log_debug(f"\nCurrently considering {test.name}")
        solver.add(test.__z3__ == z3.And(*test.query()))
        # import pdb; pdb.set_trace()
        solver.add(test.__z3__ == z3.BoolVal(True))

        # Check the model 
        result = solver.check()
        if result == z3.unsat:
            tea_logger.log_debug("Test is unsat.\n")
            solver.pop() 
        elif result == z3.unknown:
            print("failed to solve")
            try:
                pass
            except z3.Z3Exception:
                return
        else:
            model = solver.model()
            test_invalid = False
            val = False
            # Does the test apply?
            # Would this ever be false??
            if model and z3.is_true(model.evaluate(test.__z3__)):
                # Verify the properties for that test
                for prop in test._properties:
                    if is_assumed_prop(assumed_props, prop):
                        tea_logger.log_debug(f"User asserted property: {prop._name}.")
                        val = True
                        solver.add(prop.__z3__ == z3.BoolVal(val))
                    else: 
                        tea_logger.log_debug(f"Testing assumption: {prop._name}.")
                    
                        # Does this property need to hold for the test to be valid?
                        # If so, verify that the property does hold
                        if model and z3.is_true(model.evaluate(prop.__z3__)):
                            val = verify_prop(dataset, combined_data, prop)
                            if val: 
                                tea_logger.log_debug(f"Property holds.")
                            else: # The property does not verify
                                assert (val == False)
                                tea_logger.log_debug(f"Property FAILS")
                                # if not test_invalid: 
                                solver.pop() # remove the last test
                                test_invalid = True
                                model = None
                                # else: # test is already invalid. Going here just for completeness of logging
                                #     tea_logger.log_debug(f"EVER GET HERE?")
                            solver.add(prop.__z3__ == z3.BoolVal(val))
        solver.push() # Push latest state as backtracking point


        
        
    solver.check()
    model = solver.model() # final model
    tests_to_conduct = []
    # Could add all the test props first 
    # Then add all the tests 
    for test in all_tests():
        if model and z3.is_true(model.evaluate(test.__z3__)):
            tests_to_conduct.append(test.name)
        elif not model: # No test applies
            pass

    # reset_all_tests()
    return tests_to_conduct


def which_props(tests_names: list, var_names: List[str]):
    stat_vars: List[StatVar] = []
    for var_name in var_names:
        stat_vars.append(StatVar(var_name))

    axioms = construct_axioms(stat_vars)

    tests: List[StatisticalTest] = []
    for test_name in tests_names:
        for test in all_tests():
            if test.name == test_name:
                test.apply(*stat_vars)
                test._populate_properties() 
                tests.append(test)

    test_queries = []
    important_soft_constraints = []
    for test in tests:
        # The Bool representing whether the test can be run with the given property
        # values is a hard constraint because we want to satisfy all tests, but may
        # not be able to.
        important_soft_constraints.append(test.__z3__)

        # The relationship of the properties of variables required to run a specific
        # test is an axiom and cannot be violated.
        axioms.append(test.__z3__ == z3.And(*test.query()))

        # test_queries contains the individual property constraints (e.g. continuous(x)).
        # These will all be soft constraints with low weight because they may need to be
        # violated.
        test_queries += test.query()

    s = z3.Optimize()

    # Set weight of tests to large number so they are satisfied over properties.
    # print("\nImportant soft constraints:")
    for constraint in important_soft_constraints:
        # print(constraint)
        s.add_soft(constraint, 1000)

    # Prefer to violate properties by giving them low weight.
    # print("\nSoft constraints:")
    for soft_constraint in test_queries:
        # print(soft_constraint)
        s.add_soft(soft_constraint, 1)

    # Axioms cannot be violated.
    # print("\nHard constraints:")
    for axiom in axioms:
        # print(axiom)
        s.add(axiom)

    result = s.check()
    if result == z3.unsat:
        # import pdb; pdb.set_trace()
        print("no solution")
    elif result == z3.unknown:
        print("failed to solve")
        try:
            # print(s.model())
            print(f"Failed to solve: {s.model()}")
        except z3.Z3Exception:
            return
    else:
        model = s.model()
        _tests_and_properties = {}
        _test_to_broken_properties = {}
        for test in tests:
            test_name = test.name
            _tests_and_properties[test_name] = {}
            _test_to_broken_properties[test_name] = []

            # TODO: Unprotect test._properties.
            for test_property in test._properties:
                property_identifier = ""# test_property._name
                for test_var in test_property.test_vars:
                    property_identifier += "variable %s : " % test_var.name
                property_identifier += test_property._name
                property_result = bool(model.evaluate(test_property.__z3__))
                _tests_and_properties[test_name][property_identifier] = property_result
                if not property_result:
                    _test_to_broken_properties[test_name].append(property_identifier)

        return _tests_and_properties, _test_to_broken_properties


# Contains all the global values referenced throughout Tea

# For study design dictionary
btw_subj = 'between subjects'
within_subj = 'within subjects'
uid = 'key'

study_type_identifier = 'study type'
experiment_identifier = 'experiment'
observational_identifier = 'observational study'
iv_identifier = 'independent variables'
dv_identifier = 'dependent variables'
null_identifier = 'variables'
outcome_identifier = 'outcome variables'
contributor_identifier = 'contributor variables'
#quasi_experiment = 'quasi_experiment'

# For statistical properties of data
normal_distribution = 'normal distribution'
groups_normal = 'groups normally distributed'
log_normal_distribution = 'log normal distribution'
variance = 'variance'
sample_size = 'sample size'
num_categories = 'number of categories'
eq_variance = 'equal variance'
paired = 'paired'
cat_distribution = 'category distributions'

# For non-statistical meta-properties of data
name = 'var_name'
data_type = 'dtype'
categories = 'categories'
query = 'query'

# For solver
# Maps assumption names (from user) to property names
alpha_keywords = ['Type I (False Positive) Error Rate', 'alpha']

assumptions_to_properties = {
    normal_distribution : ['is_normal'],
    groups_normal : ['is_groups_normal'],
    log_normal_distribution : ['is_log_normal'],
    eq_variance : ['has_equal_variance']
}

# For solver, how to treat user assumptions
# MODE = 'strict' #can be 'strict' or 'relaxed'


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tea',
 'tea.helpers',
 'tea.helpers.constants',
 'tea.logging',
 'tea.runtimeDataStructures',
 'tea.z3_solver']

package_data = \
{'': ['*']}

install_requires = \
['attrs==21.4.0',
 'bootstrapped==0.0.2',
 'pandas==1.3.5',
 'parameterized==0.8.1',
 'requests==2.27.1',
 'scikit-learn==1.0.2',
 'scipy==1.7.3',
 'statsmodels>=0.12.2,<0.13.0',
 'sympy==1.9',
 'urllib3==1.26.8',
 'z3-solver==4.8.14.0']

setup_kwargs = {
    'name': 'tealang',
    'version': '0.4.1',
    'description': 'Tea: A High-level Language and Runtime System to Automate Statistical Analysis',
    'long_description': None,
    'author': 'Eunice Jun',
    'author_email': 'eunice.m.jun@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7.11,<3.11',
}


setup(**setup_kwargs)
